@@ -51,12 +51,16 @@
 #include <string.h>
 #include <ctype.h>
 #include <time.h>
-#include <linux/ax25.h>
-#include <linux/rose.h>
 #include <signal.h>
 #include <errno.h>
-#include <netax25/axconfig.h>
+#include <netax25/kernel_ax25.h>
+#include <netax25/kernel_rose.h>
+#include <netax25/ax25.h>
 #include <netax25/axlib.h>
+#include <netax25/axconfig.h>
+#ifndef AX25_MTU
+#define AX25_MTU 256
+#endif
 #include <netdb.h>
 
 #include "mic_e.h"
@@ -66,7 +70,7 @@
 #define PACKAGE "aprsdigi"
 #endif
 #ifndef VERSION
-#define VERSION "$Revision: 1.42 $";
+#define VERSION "$Revision: 1.1.1.1 $";
 #endif
 
 /* some defines that really belong in a header file! */
@@ -117,9 +121,24 @@ struct budlist {		/* a given budlist */
 
 static struct budlist *Bl_head = NULL, *Bl_last = NULL;
 
+struct dupelist_entry {		/* dupelist entry */
+  struct dupelist_entry *de_next;
+  char *de_dev;			/* interface device name to dupe to */
+};
+
+struct dupelist {		/* a given dupelist */
+  struct dupelist *dl_next;
+  int dl_list_no;		/* list number */
+  struct interface_list *dl_il;	/* list of interfaces to dupe to */
+  struct dupelist_entry *dl_ent_head, *dl_ent_last;
+};
+
+static struct dupelist *Dl_head = NULL, *Dl_last=NULL;
+
 static struct interface {
   char *port;			/* axports port name */
   char *dev;			/* kernel device name */
+  char *devname;		/* user-friendly device name */
   char *tag;			/* optional tag text */
   int taglen;
   int idinterval;		/* seconds between IDs; 0 for none. */
@@ -143,14 +162,15 @@ static struct interface {
   int idlen;
   struct statistics stats;	/* statistics */
   struct budlist *bud;		/* budlist */
+  struct dupelist *dupe;	/* list of interfaces to dupe to */
 } Intf[MAXINTF];
 #define MYCALL(n) Intf[n].aliases[0]
 #define I_MYCALL(i) i->aliases[0]
 static int N_intf = 0;
 
 struct interface_list {		/* a list of interfaces (duh) */
-  struct interface *i;
   struct interface_list *next;
+  struct interface *i;
 };
 
 struct stuff {			/* maybe I should learn C++... */
@@ -169,6 +189,7 @@ static int Testing = 0;
 static int Digi_SSID = 0;
 static int Kill_dupes = 0;	/* kill dupes even in conventional mode */
 static int Kill_loops = 0;	/* kill loops */
+static int Doing_dupes = 0;	/* dupelist was set on some interface */
 static char *Logfile = NULL;
 static ax25_address Aprscall;	/* replace mic-e encoded to-call with this */
 static ax25_address Trace_dummy; /* dummy call for tracen-n substitution */
@@ -249,7 +270,7 @@ static void print_it(FILE *,struct ax_calls *,unsigned char *,int len);
 static void add_text(u_char **,int *,u_char *,int,char *,int);
 static int dupe_packet(struct ax_calls *,u_char *,int);
 static int loop_packet(struct ax_calls *,u_char *,int);
-static int xmit(struct stuff *s);
+static int xmit(struct stuff *s, struct interface_list *list);
 static void set_id(void);
 static void rx_packet(struct interface *i, u_char *buffer, int len);
 static void fix_recvfrom(struct stuff *s);
@@ -260,6 +281,10 @@ static void budlist_add(char *, int);
 static int budlist_permit(struct stuff *);
 static void budlist_print_all(FILE *f);
 static void budlist_print(FILE *f,struct budlist *);
+static void dupelist_add(char *);
+static void dupelist_print_all(FILE *f);
+static void dupelist_print(FILE *f,struct dupelist *);
+static void dupelist_init(void);
 static void printaddr(struct sockaddr_storage *s, char *buf, int buflen);
 static void setmask(int, u_char *);
 static int cmpmask(int, u_char *, u_char *, u_char *);
@@ -354,6 +379,7 @@ rx_loop()
 
 /* Parse a received packet, convert and digipeat if necessary. */
 
+static void rx_dupeit(struct stuff *s);
 static int rx_nodigi(struct stuff *s,int r);
 static int rx_dupe(struct stuff *s);
 static int rx_to_me(struct stuff *s);
@@ -396,6 +422,8 @@ rx_packet(struct interface *i,	/* which interface received on */
 
   /* go through a bunch of choices until we eat the packet or die:-) */
 
+  if (Doing_dupes)
+    rx_dupeit(&s);		/* blindly dupe it */
   if (rx_bud_deny(&s))		/* budlisted */
     return;
   else if (rx_from_me(&s))      /* don't digi my own repeated beacons! */
@@ -436,6 +464,22 @@ fix_recvfrom(struct stuff *s)
   }
 }
 
+static void
+rx_dupeit(struct stuff *s)
+{
+  int n;
+
+  if (s->i->dupe && s->i->dupe->dl_il) {
+    if (Verbose) {
+      fprintf(stderr,"%s: duping to dupelist %d\n",s->i->devname,
+	      s->i->dupe->dl_list_no);
+    }
+    s->out = s->in;
+    if (xmit(s,s->i->dupe->dl_il) < 0) 
+      perror("xmit");
+    }
+}
+
 static int
 rx_nodigi(struct stuff *s,int r)
 {
@@ -686,7 +730,7 @@ rx_ssid(struct stuff *s)
       }
     }
     ++s->i->stats.ssid;
-    if (xmit(s) < 0)
+    if (xmit(s,NULL) < 0)
       perror("xmit");
     result = 1;
   }
@@ -748,7 +792,7 @@ rx_flood(struct stuff *s)
 	      ax25_ntoa_pretty(&s->out.ax_digi_call[s->out.ax_next_digi]));
     }
     ++s->i->stats.flood;
-    if (xmit(s) < 0) 
+    if (xmit(s,NULL) < 0) 
       perror("xmit");
     return 1;
   }
@@ -775,7 +819,7 @@ rx_digi(struct stuff *s)
       s->out = s->in;	/* copy input list to output unmodifed */
       s->out.ax_digi_call[s->out.ax_next_digi].ax25_call[ALEN] |= REPEATED;
       ++s->i->stats.digi;
-      if (xmit(s) < 0) 
+      if (xmit(s,NULL) < 0) 
 	perror("xmit");
       result  = 1;
   }
@@ -904,9 +948,11 @@ reformat(struct stuff *s,	/* "class data":-) */
  * xmit looks at the next digipeater call to see which interfaces the
  * packet has to go out.  By giving several interfaces the same callsign
  * or alias, multi-way fanout and/or gatewaying can be performed.
+ *
+ * An overriding interface list passed in is used by rx_dupeit().
  */
 static int
-xmit(struct stuff *s)
+xmit(struct stuff *s, struct interface_list *dupelist)
 {
   struct callsign_list *c;
   struct interface_list *l;
@@ -914,12 +960,14 @@ xmit(struct stuff *s)
   int thisflags = 0;
 
   /* transmit the packet on the interface(s) of the next callsign */
-  if ((c = calltab_entry(&s->out.ax_digi_call[s->out.ax_next_digi],&thisflags)) == NULL) {
-    fprintf(stderr,"xmit: Assertion failed: call %s not found\n",
-	    ax25_ntoa_pretty(&s->out.ax_digi_call[s->out.ax_next_digi]));
-    return -1;
+  if (!dupelist) {
+    if ((c = calltab_entry(&s->out.ax_digi_call[s->out.ax_next_digi],&thisflags)) == NULL) {
+      fprintf(stderr,"xmit: Assertion failed: call %s not found\n",
+	      ax25_ntoa_pretty(&s->out.ax_digi_call[s->out.ax_next_digi]));
+      return -1;
+    }
   }
-  for(l = c->l; l; l = l->next) { /* loop over each tx interface */
+  for(l = (dupelist)?dupelist:c->l; l; l = l->next) { /* loop over each tx interface */
     u_char obuf[AX25_MTU];
     u_char *op = obuf;
     int oleft = sizeof(obuf);
@@ -930,12 +978,12 @@ xmit(struct stuff *s)
 
     if (!(l->i->i_flags & I_TX)) {
       if (Verbose)
-	fprintf(stderr,"%s: interface will not xmit.\n",l->i->dev);
+	fprintf(stderr,"%s: interface will not xmit.\n",l->i->devname);
       continue;
     }
     if (!(l->i->i_flags & I_TXSAME) && l->i == s->i) {
       if (Verbose)
-	fprintf(stderr,"%s: will not retransmit on same interface.\n",l->i->dev);
+	fprintf(stderr,"%s: will not retransmit on same interface.\n",l->i->devname);
       continue;
     }
 
@@ -964,15 +1012,17 @@ xmit(struct stuff *s)
        * rx_flood will have inserted tracecall in front of TRACEn-n.
        * but MYCALL subst happens here since it is iface-specific.
        */
-      if (calls.ax_digi_call[calls.ax_next_digi].ax25_call[ALEN]&REPEATED) {
-	if ((!(thisflags&C_IS_FLOOD) && l->i->i_flags&SUBST_MYCALL)
-	    || ((thisflags&C_IS_FLOOD) && !(thisflags&C_IS_FLOODN))
-	    || ((thisflags&(C_IS_FLOOD|C_IS_FLOODN|C_IS_TRACE))
-		== (C_IS_FLOOD|C_IS_FLOODN|C_IS_TRACE))) {
-	  calls.ax_digi_call[calls.ax_next_digi] = I_MYCALL(l->i);
-	  calls.ax_digi_call[calls.ax_next_digi].ax25_call[ALEN] |= REPEATED;
+      if (!dupelist) {		/* if dupeing, ignore this crap */
+	if (calls.ax_digi_call[calls.ax_next_digi].ax25_call[ALEN]&REPEATED) {
+	  if ((!(thisflags&C_IS_FLOOD) && l->i->i_flags&SUBST_MYCALL)
+	      || ((thisflags&C_IS_FLOOD) && !(thisflags&C_IS_FLOODN))
+	      || ((thisflags&(C_IS_FLOOD|C_IS_FLOODN|C_IS_TRACE))
+		  == (C_IS_FLOOD|C_IS_FLOODN|C_IS_TRACE))) {
+	    calls.ax_digi_call[calls.ax_next_digi] = I_MYCALL(l->i);
+	    calls.ax_digi_call[calls.ax_next_digi].ax25_call[ALEN] |= REPEATED;
+	  }
+	  ++calls.ax_next_digi;
 	}
-	++calls.ax_next_digi;
       }
       /* XXX - deal with detunneled 3rdparty (empty digi list) -- or is
 	 some default digi list supposed to be used? */
@@ -990,7 +1040,7 @@ xmit(struct stuff *s)
       add_text(&op,&oleft,vecp[n],vecl[n],l->i->tag,l->i->taglen); /* fill in the info field */	
 
       olen = sizeof(obuf) - oleft;
-      if (Logfile) {
+      if (Logfile) {		/* XXX this gets duplicated crap for multiple intfs */
 	FILE *of;
 	if (Logfile[0] == '-' && Logfile[1] == '\0')
 	  of = stdout;
@@ -1266,7 +1316,7 @@ xmit_id(struct interface *i)
   }
   if (sendto(i->tsock,i->idbuf,i->idlen,
 	     0,(struct sockaddr *)&i->tsa,i->tsa_size) < 0) /* ship it off */
-    perror(i->dev);
+    perror(i->devname);
   i->i_flags &= ~I_NEED_ID;
   i->next_id = 0;
   ++i->stats.ids;
@@ -1333,14 +1383,61 @@ identify_final(int sig)		/* wake up and ID! */
   }
 }
 
+#include <getopt.h>
+enum {NONE=no_argument,REQD=required_argument,OPT=optional_argument};
+static struct option opts[] = {
+  {"verbose",OPT,0,'v'},
+  {"testing",NONE,0,'T'},
+  {"interface",REQD,0,'p'},
+  {"port",REQD,0,'p'},
+  {"trace",REQD,0,'F'},
+  {"flood",REQD,0,'f'},
+  {"east",REQD,0,'e'},
+  {"west",REQD,0,'w'},
+  {"north",REQD,0,'n'},
+  {"south",REQD,0,'s'},
+  {"digipath",REQD,0,'d'},
+  {"tag",REQD,0,'t'},
+  {"keep",REQD,0,'k'},
+  {"logfile",REQD,0,'l'},
+  {"idinterval",REQD,0,'i'},
+  {"subst_mycall",NONE,0,'C'},
+  {"nosubst_mycall",NONE,0,'c'},
+  {"mice_xlate",NONE,0,'M'},
+  {"nomice_xlate",NONE,0,'m'},
+  {"x1j4_xlate",NONE,0,'X'},
+  {"nox1j4_xlate",NONE,0,'x'},
+  {"3rdparty",NONE,0,'3'},
+  {"no3rdparty",NONE,0,'0'},
+  {"kill_dupes",NONE,0,'D'},
+  {"kill_loops",NONE,0,'L'},
+  {"bud",REQD,0,'B'},
+  {"nobud",REQD,0,'b'},
+  {"version",NONE,0,'V'},
+  {"help",NONE,0,'?'},
+  {"opts",REQD,0,'o'},  /* ran out of letters */
+#define O(x) ((int)(256+(int)x))
+  {"rx",NONE,0,O('R')},
+  {"norx",NONE,0,O('r')},
+  {"tx",NONE,0,O('T')},
+  {"notx",NONE,0,O('t')},
+  {"txsame",NONE,0,O('S')},
+  {"notxsame",NONE,0,O('s')},
+  {"dupe",REQD,0,O('d')},
+  {"duplicate",REQD,0,O('d')},
+  {0,0,0,0}
+};
+static char *optstring = "CcMmXxF:f:n:s:e:w:t:k:l:i:d:p:DLVvT30o:B:b:";
+
 static void
 do_opts(int argc, char **argv)
 {
-  char s;
+  int s;
+  int opt_index = 0;
 
   /* set interface defaults */
   I_flags = I_TX|I_RX|I_TXSAME;
-  while ((s = getopt(argc, argv, "CcMmXxF:f:n:s:e:w:t:k:l:i:d:p:DLVvT30o:B:b:")) != -1) {
+  while ((s = getopt_long(argc, argv, optstring, opts, &opt_index)) != -1) {
     int p = -1;			/* used for NSEW hack, below */
     int fflags = 0;		/* use for fF hack */
     switch (s) {
@@ -1365,6 +1462,7 @@ do_opts(int argc, char **argv)
 	i->taglen = Taglen;
 	i->idinterval = Idinterval;
 	i->bud = Bl_last;
+	i->dupe = Dl_last;
 	if (strncasecmp(optarg,"udp:",4) == 0) { /* IP/UDP port */
 	  i->i_flags |= I_COOKED;
 	  i->i_proto = P_UDP;
@@ -1411,8 +1509,10 @@ do_opts(int argc, char **argv)
 	    i->n_aliases--;	/* alias 0 is MYCALL */
 	}
       }
-      /* -p is an implicit -B - */
+#ifdef notdef
+      /* -p is an implicit -B - XXX why? */
       budlist_add("-",0);
+#endif
       break;
     case 'F':			/* flooding trace alias */
       fflags = C_IS_TRACE;
@@ -1521,6 +1621,27 @@ do_opts(int argc, char **argv)
 	usage();
       }
       break;
+    case O('d'):		/* dupe to interface w/no modification */
+      dupelist_add(optarg);
+      break;
+    case O('R'):
+      I_flags |= I_RX;
+      break;
+    case O('r'):
+      I_flags &= ~I_RX;
+      break;
+    case O('T'):
+      I_flags |= I_TX;		/* TX enable */
+      break;
+    case O('t'):
+      I_flags &= ~I_TX;	/* TX disable */
+      break;
+    case O('S'):
+      I_flags |= I_TXSAME;	/* re-TX on RX interface */
+      break;
+    case O('s'):
+      I_flags &= ~I_TXSAME;	/* no re-TX on RX interface */
+      break;
     case 'D':
       Kill_dupes++;
       break;
@@ -1559,89 +1680,101 @@ usage()
 {
   fprintf(stderr,"Usage: aprsdigi [-CcDLMXv] [-n|s|e|w path] [-fF call] [-t tag] [-k secs] [-l logfile] [-i interval] [-o r|R|t|T|s|S] [-p port:alias1,alias2...] [-B budlist] [-b budlist] ...\n");
   fprintf(stderr," general options:\n");
-  fprintf(stderr," -v       -- produce verbose debugging output\n");
-  fprintf(stderr," -T       -- test mode: listen to my packets too\n");
-  fprintf(stderr," -D       -- suppress Duplicate packets\n");
-  fprintf(stderr," -L       -- suppress Looping packets\n");
-  fprintf(stderr," -V       -- print program version and exit\n");
+  fprintf(stderr," -v | --verbose     -- produce verbose debugging output\n");
+  fprintf(stderr," -T | --testing     -- test mode: listen to my packets too\n");
+  fprintf(stderr," -D | --kill_dupes  -- suppress Duplicate packets\n");
+  fprintf(stderr," -L | --kill_loops  -- suppress Looping packets\n");
+  fprintf(stderr," -V | --version     -- print program version and exit\n");
   fprintf(stderr," -n|s|e|w -- set North|South|East|West SSID directional path\n");
-  fprintf(stderr," -f       -- set Flooding callsign (usually WIDE)\n");
-  fprintf(stderr," -F       -- set Flooding TRACE callsign (usually TRACE)\n");
-  fprintf(stderr," -k       -- seconds of old dupes to remember\n");
-  fprintf(stderr," -l       -- log digipeated packets here\n");
+  fprintf(stderr," --north | --south | --east | --west\n");
+  fprintf(stderr," -f | --flood       -- set Flooding callsign (usually WIDE)\n");
+  fprintf(stderr," -F | --trace       -- set Flooding TRACE callsign (usually TRACE)\n");
+  fprintf(stderr," -k | --keep        -- seconds of old dupes to remember\n");
+  fprintf(stderr," -l | --logfile     -- log digipeated packets here\n");
   fprintf(stderr," per-interface options (put *before* each -p as needed):\n");
-  fprintf(stderr," -C (-c)  -- do (not) perform Mycall substitution\n");
-  fprintf(stderr," -M (-m)  -- do (not) perform Mic-E translation\n");
-  fprintf(stderr," -X (-x)  -- do (not) perform X1J4 translation\n");
-  fprintf(stderr," -3       -- perform 3rd party tunneling\n");
-  fprintf(stderr," -0       -- perform transparent tunneling (default)\n");
-  fprintf(stderr," -o r (R) -- do (not) mark interface receive-only\n");
-  fprintf(stderr," -o t (T) -- do (not) mark interface transmit-only\n");
-  fprintf(stderr," -o s (S) -- do (not) supress retransmission on RX intf.\n");
-  fprintf(stderr," -i       -- seconds between ID transmissions\n");
+  fprintf(stderr," -C (-c) | --[no]subst_mycall  -- do (not) perform Mycall substitution\n");
+  fprintf(stderr," -M (-m) | --[no]mice_xlate    -- do (not) perform Mic-E translation\n");
+  fprintf(stderr," -X (-x) | --[no]x1j4_xlate    -- do (not) perform X1J4 translation\n");
+  fprintf(stderr," -3 | --3rdparty     -- perform 3rd party tunneling\n");
+  fprintf(stderr," -0 | --no3rdparty   -- perform transparent tunneling (default)\n");
+  fprintf(stderr," -o r (R) | --[no]rx           -- do (not) mark interface receive-only\n");
+  fprintf(stderr," -o t (T) | --[no]tx           -- do (not) mark interface transmit-only\n");
+  fprintf(stderr," -o s (S) | --[no]txsame       -- do (not) supress retransmission on RX intf.\n");
+  fprintf(stderr," -i | --idinterval   -- seconds between ID transmissions\n");
   fprintf(stderr,"             (Use \"-i 0\" to disable IDs)\n");
-  fprintf(stderr," -t       -- default tag text to add to received packets\n");
+  fprintf(stderr," -t | --tag          -- tag text to add to received packets\n");
   fprintf(stderr,"             (Use \"-t -\" to reset to empty)\n");
-  fprintf(stderr," -B call|addr -- accept traffic only from matching source.\n");
-  fprintf(stderr," -b call|addr -- deny traffic from matching source.\n");
+  fprintf(stderr," -B | --bud call|addr -- accept traffic only from matching source.\n");
+  fprintf(stderr," -b | --nobud call|addr -- deny traffic from matching source.\n");
+  fprintf(stderr," --dupe interface     -- copy traffic from this interface.\n");
   fprintf(stderr,"\n");
-  fprintf(stderr," -p ax25:port:a1,a2,... -- port name and optional list of aliases\n");
+  fprintf(stderr," -p | --port | --interface ax25:port:a1,a2,... -- port name and optional list of aliases\n");
   fprintf(stderr,"             port is from /etc/ax25/axports\n");
   fprintf(stderr,"             (primary callsign is obtained from the port)\n");
-  fprintf(stderr," -p udp:address/port/ttl:call,a1,a2... -- IP host,service,ttl,call and optional list of aliases\n");
-  fprintf(stderr," -p unix:path:call,a1,a2... -- FIFO filename, call and optional list of aliases\n");
+  fprintf(stderr," -p | --port | --interface udp:address/port/ttl:call,a1,a2... -- IP host,service,ttl,call and optional list of aliases\n");
+  fprintf(stderr," -p | --port | --interface unix:path:call,a1,a2... -- FIFO filename, call and optional list of aliases\n");
   exit(1);
 }
 
 static void
 usage_udp()
 {
-  fprintf(stderr,"aprsdigi: Use -p udp:host/service/ttl:callsign,a1,a2,...\n");
-  fprintf(stderr," example: -p udp:224.1.2.3/1234/127:N2YGK-2,RELAY,WIDE\n");
-  fprintf(stderr,"          -p udp:192.168.1.2/1234/16:N2YGK-2,RELAY,WIDE\n");
-  fprintf(stderr,"          -p udp:n0clu.ampr.net/1234/16:N2YGK-2,RELAY,WIDE\n");
-  fprintf(stderr,"          -p udp:2002:123:45::ab:cd/1234/16:N2YGK-2,RELAY,WIDE\n");
+  fprintf(stderr,"aprsdigi: Use --int udp:host/service/ttl:callsign,a1,a2,...\n");
+  fprintf(stderr," example: --int udp:224.1.2.3/1234/127:N2YGK-2,RELAY,WIDE\n");
+  fprintf(stderr,"          --int udp:192.168.1.2/1234/16:N2YGK-2,RELAY,WIDE\n");
+  fprintf(stderr,"          --int udp:n0clu.ampr.net/1234/16:N2YGK-2,RELAY,WIDE\n");
+  fprintf(stderr,"          --int udp:2002:123:45::ab:cd/1234/16:N2YGK-2,RELAY,WIDE\n");
   exit(1);
 }
 
 static void
 usage_ax25()
 {
-  fprintf(stderr,"aprsdigi: Use -p ax25:port:a1,a2,...\n");
-  fprintf(stderr," example: -p ax25:1:N2YGK-7,RELAY,WIDE\n");
+  fprintf(stderr,"aprsdigi: Use --int ax25:port:a1,a2,...\n");
+  fprintf(stderr," example: --int ax25:1:N2YGK-7,RELAY,WIDE\n");
   exit(1);
 }
 
 static void
 usage_unix()
 {
-  fprintf(stderr,"aprsdigi: Use -p unix:path:a1,a2,...\n");
-  fprintf(stderr," example: -p unix:/tmp/fifo:N2YGK-7,RELAY,WIDE\n");
+  fprintf(stderr,"aprsdigi: Use --int unix:path:a1,a2,...\n");
+  fprintf(stderr," example: --int unix:/tmp/fifo:N2YGK-7,RELAY,WIDE\n");
   exit(1);
 }
 
 static void
 usage_budlist()
 {
-  fprintf(stderr,"aprsdigi: Use -B|b ax25:callsign[-SSID]|ip:host|udp:addr[/mask].\n");
-  fprintf(stderr,"          -B permits.  -b denies.\n");
-  fprintf(stderr,"          \"-B -\" resets the list to empty\n");
+  fprintf(stderr,"aprsdigi: Use --[no]bud ax25:callsign[-SSID]|ip:host|udp:addr[/mask].\n");
+  fprintf(stderr,"          \"--bud -\" resets the list to empty\n");
   fprintf(stderr,"          Use callsign-SSID to match specific station.\n");
   fprintf(stderr,"          Use callsign to match all SSIDs.\n");
   fprintf(stderr,"          Use host to match a given IP host.\n");
   fprintf(stderr,"          Use addr/mask to match a given IP range.\n");
-  fprintf(stderr,"examples: -b ax25:n0clu-14  -- matches only n0clu-14.\n");
-  fprintf(stderr,"          -b ax25:n0clu     -- matches all n0clu's\n");
-  fprintf(stderr,"          -b ip:n0clu.ampr.net  -- matches one host.\n");
-  fprintf(stderr,"          -b ip:192.168.5.3     -- matches one host.\n");
-  fprintf(stderr,"          -b ip:192.168.5.3/32  -- matches one host.\n");
-  fprintf(stderr,"          -b ip:192.168.5.0/25  -- matches subnet\n");
-  fprintf(stderr,"          -b ip:fe80::201:3ff:fe9a:38c6  -- matches one host.\n");
-  fprintf(stderr,"          -b ip:fe80::201:3ff:fe9a:38c6/128  -- matches one host.\n");
-  fprintf(stderr,"          -b ip:2002:905::/32  -- matches prefix.\n");
+  fprintf(stderr,"examples: --bud ax25:n0clu-14  -- matches only n0clu-14.\n");
+  fprintf(stderr,"          --bud ax25:n0clu     -- matches all n0clu's\n");
+  fprintf(stderr,"          --bud ip:n0clu.ampr.net  -- matches one host.\n");
+  fprintf(stderr,"          --bud ip:192.168.5.3     -- matches one host.\n");
+  fprintf(stderr,"          --bud ip:192.168.5.3/32  -- matches one host.\n");
+  fprintf(stderr,"          --bud ip:192.168.5.0/25  -- matches subnet\n");
+  fprintf(stderr,"          --bud ip:fe80::201:3ff:fe9a:38c6  -- matches one host.\n");
+  fprintf(stderr,"          --bud ip:fe80::201:3ff:fe9a:38c6/128  -- matches one host.\n");
+  fprintf(stderr,"          --bud ip:2002:905::/32  -- matches prefix.\n");
   exit(1);
 }
 
+static void
+usage_dupelist()
+{
+  fprintf(stderr,"aprsdigi: Use --dupe ax25:interface|udp:addr|unix:file.\n");
+  fprintf(stderr,"          \"--dupe -\" resets the list to empty\n");
+  exit(1);
+}
+
+/*
+ * do_ports: Initialize interfaces.
+ */
 static void
 do_ports()
 {
@@ -1673,6 +1806,7 @@ do_ports()
     }
   }
   calltab_init();
+  dupelist_init();
 }
 
 static void
@@ -1685,6 +1819,7 @@ do_port_ax25(struct interface *i)
     fprintf(stderr, "aprsdigi: invalid ax25 port name - %s\n", i->port);
     exit(1);
   }
+  i->devname = strdup(i->port);
   i->rsa_size = i->tsa_size = sizeof(struct sockaddr_ax25);
 
   if (!(i->i_flags & I_TX)) {
@@ -1706,14 +1841,14 @@ do_port_ax25(struct interface *i)
     strcpy(((struct sockaddr *)&i->rsa)->sa_data, i->dev);
     ((struct sockaddr *)&i->rsa)->sa_family = AF_AX25;
     if (bind(i->rsock, (struct sockaddr *)&i->rsa, i->rsa_size) < 0) {
-      die(i->dev);
+      die(i->devname);
     }
   }
 
   /* set mycall (zeroth alias) to the port's */
   strcpy(ifr.ifr_name, i->dev); /* get this port's callsign */
   if (ioctl(i->rsock, SIOCGIFHWADDR, &ifr) < 0) {
-    die(i->dev);
+    die(i->devname);
   }
   if (ifr.ifr_hwaddr.sa_family != AF_AX25) {
     fprintf(stderr,"%s: not AX25\n",i->port);
@@ -1733,17 +1868,19 @@ do_port_udp(struct interface *i)
   int ttlval;
   struct addrinfo *ai, hints = {0, PF_UNSPEC, SOCK_DGRAM, 0};
 
-  i->dev = i->port;		/* might as well just use the same thing */
   /* hostname/service(port)/ttl */
+  i->dev = i->port;
   name = i->port;
   service = strchr(name,'/');
   if (!service || *service != '/')
     usage_udp();
-  *service++ = '\0';
-  ttl = strchr(service,'/');
+  //  *service++ = '\0';
+  ttl = strchr(&service[1],'/');
   if (!ttl || *ttl != '/')
     usage_udp();
   *ttl++ = '\0';
+  i->devname = strdup(name);
+  *service++ = '\0';
   if ((ttlval = atoi(ttl)) <= 0)
       usage_udp();
 
@@ -1785,7 +1922,7 @@ do_port_udp(struct interface *i)
     bzero(&i->tsa,sizeof(i->tsa));
     bcopy((struct sockaddr*)ai->ai_addr,&i->tsa,sizeof(i->tsa)); /* fill sockaddr w/sockaddr_in */
     if (connect(i->tsock, (struct sockaddr *)&i->tsa, i->tsa_size) < 0) {
-      die(i->dev);
+      die(i->devname);
     }
   }
 
@@ -1811,7 +1948,7 @@ do_port_udp(struct interface *i)
     bzero(&i->rsa,sizeof(i->rsa));
     bcopy((struct sockaddr*)ai->ai_addr,&i->rsa,sizeof(i->rsa));
     if (bind(i->rsock, (struct sockaddr *)&i->rsa, i->rsa_size) < 0) {
-      die(i->dev);
+      die(i->devname);
     }
     /* if the UDP socket is a multicast group, then do IGMP join for rsock */
     
@@ -1837,7 +1974,7 @@ do_port_unix(struct interface *i)
   char name[sizeof(sun->sun_path)];
   
   i->dev = i->port;
-
+  i->devname = strdup(i->dev);
   strncpy(name,i->dev,sizeof(name));
   i->rsa_size = i->tsa_size = sizeof(struct sockaddr_un);
   /* the receive socket */
@@ -1929,13 +2066,18 @@ floodtype(ax25_address *c, int i)
     return C_IS_NOFLOOD;
 }
 
+#define DEVTYPE(i) ((i)->i_proto == P_UDP)?"udp":\
+	    ((i)->i_proto == P_AX25)?"ax25":\
+	    ((i)->i_proto == P_UNIX)?"unix":"???"
+
+
 static void
 check_config()
 {
   int n, j;
   struct callsign_list *cl;
   
-  printf("Linux APRS(tm) digipeater\nCopyright (c) 1996,1997,1999,2001,2002,2003 Alan Crosswell, n2ygk@weca.org\n");
+  printf("Linux APRS(tm) digipeater\nCopyright (c) 1996,1997,1999,2001,2002,2003,2004 Alan Crosswell, n2ygk@weca.org\n");
   printf("Version: aprsdigi %s-%s\n",PACKAGE,VERSION);
   printf("This is free software covered under the GNU Public License.\n");
   printf("There is no warranty.  See the file COPYING for details.\n\n");
@@ -1952,7 +2094,7 @@ check_config()
       strcat(t,"n-n");		/* indicate a WIDEn-n */
     printf("%-9.9s",t);
     for (l = cl->l; l; l = l->next) {
-      printf(" %-9.9s",l->i->port);
+      printf(" %s:%s",DEVTYPE(l->i),l->i->devname);
     }
     printf("\n");
   }
@@ -2009,11 +2151,9 @@ config_print(FILE *f)
 
   fprintf(f,"# configuration:\n");
   budlist_print_all(f);
+  dupelist_print_all(f);
   for (i = Intf, n = 0; n < N_intf; i++,n++) {
-    fprintf(f,"interface %s:%s\n",
-	    (i->i_proto == P_UDP)?"udp":
-	    (i->i_proto == P_AX25)?"ax25":
-	    (i->i_proto == P_UNIX)?"unix":"???",i->dev);
+    fprintf(f,"interface %s:%s\n", DEVTYPE(i),i->devname);
     fprintf(f," callsign %s\n",ax25_ntoa_pretty(&i->aliases[0]));
     for (j = 1; j < i->n_aliases; j++) {
       fprintf(f," alias %s\n",ax25_ntoa_pretty(&i->aliases[j]));
@@ -2031,6 +2171,8 @@ config_print(FILE *f)
     fprintf(f," option tag %s\n",(i->taglen)?i->tag:"#(none)");
     if (i->bud)
       fprintf(f," budlist %d\n",i->bud->bl_list_no);
+    if (i->dupe && i->dupe->dl_ent_head)
+      fprintf(f," dupelist %d\n",i->dupe->dl_list_no);
   }
   fprintf(f,"# end of configuration\n\n");
 }
@@ -2308,6 +2450,134 @@ cmpmask(int bits, u_char *val1, u_char *mask, u_char *val2)
   return 0;
 }
 
+/* gee, this looks like a copy of budlist.  I should make a class. */
+
+static void
+dupelist_add(char *item)
+{
+  struct dupelist_entry *e;
+
+  if (!item || !*item)
+    usage_dupelist();
+  ++Doing_dupes;
+  if (!Dl_head) {		/* this is first invocation */
+    Dl_head = Dl_last = (struct dupelist *)calloc(1,sizeof(struct dupelist));
+    if (!Dl_head) 
+      die("malloc");
+    Dl_head->dl_list_no = 1;
+  }
+  if (item[0] == '-' && item[1] == '\0') { /* start a new list */
+    if (Dl_last->dl_ent_last)	{ /* if the list has entries... */
+      /* ... start a new list */
+      Dl_last->dl_next = (struct dupelist *)calloc(1,sizeof(struct dupelist));
+      if (!Dl_last->dl_next)
+	die("malloc");
+      Dl_last->dl_next->dl_list_no = Dl_last->dl_list_no+1;
+      Dl_last = Dl_last->dl_next;
+    }
+    return;
+  }
+
+  if (!Dl_last->dl_ent_head) {	/* list has no elements yet */
+    Dl_last->dl_ent_head = Dl_last->dl_ent_last = 
+      (struct dupelist_entry *)calloc(1,sizeof(struct dupelist_entry));
+    if (!Dl_last->dl_ent_head)
+      die("malloc");
+  } else if (!Dl_last->dl_ent_last->de_next) { /* alloc a new entry */
+    Dl_last->dl_ent_last->de_next = 
+      (struct dupelist_entry *)calloc(1,sizeof(struct dupelist_entry));
+    if (!Dl_last->dl_ent_last->de_next)
+      die("malloc");
+    Dl_last->dl_ent_last = Dl_last->dl_ent_last->de_next;
+  }
+  /* done with all that allocation foolishness */
+  e = Dl_last->dl_ent_last;
+  e->de_dev = item;
+}
+
+static void
+dupelist_print_all(FILE *f)
+{
+  struct dupelist *l;
+
+  for (l = Dl_head; l; l=l->dl_next)
+    dupelist_print(f,l);
+}
+  
+
+static void
+dupelist_print(FILE *f, struct dupelist *dupe)
+{
+  char host[1000];
+  struct dupelist_entry *l;
+
+  if (!dupe || !dupe->dl_ent_head) 
+    return;
+  for (l = dupe->dl_ent_head; l; l = l->de_next) {
+    fprintf(f," dupelist %d %s\n",dupe->dl_list_no,l->de_dev);
+  }
+}
+
+static void add_intflist(struct interface_list **il, struct interface *i);
+
+/*
+ * dupelist_init:  Take dupelist from i->dupe, search for interfaces
+ * with matching names and add them to i->dupe->dl_il.  Do not add own
+ * interface.
+ */
+static void
+dupelist_init() 
+{
+  struct dupelist_entry *de;
+  struct dupelist *dl;
+  struct interface *i,*j;
+  int n;
+
+  for (i = Intf; i < &Intf[N_intf]; i++) {
+    if (i->dupe) {    /* walk all entries in this list */
+      for (de = i->dupe->dl_ent_head; de; de = de->de_next) { 
+	char devname[100];
+	/* search for an interface name that matches */
+	for (j = Intf, n = 0; n < N_intf; j++,n++) {
+	  sprintf(devname,"%s:%s",DEVTYPE(j),j->devname);
+	  if (Verbose)
+	    fprintf(stderr,"dupelist_init: comparing %s to %s...\n",
+		    de->de_dev,devname);
+	  if ((strcasecmp(devname,de->de_dev) == 0) && j != i) {
+	    if (Verbose)
+	      fprintf(stderr," ... matched.  Adding to interface list\n");
+	    add_intflist(&i->dupe->dl_il,j); /* add j to dupe interface list */
+	    break;
+	  }
+	}
+	if (n >= N_intf) {		/* search fell through */
+	  fprintf(stderr,"warning: dupelist %d interface %s not found.\n",
+		  i->dupe->dl_list_no,de->de_dev);
+	}
+      }
+    }
+  }
+}
+
+/*
+ * add_intflist: Add an interface to an interface list.
+ */
+static void
+add_intflist(struct interface_list **il, struct interface *i)
+{
+  struct interface_list *n = calloc(1,sizeof(struct interface_list));
+
+  if (n == NULL)
+    die("malloc");
+  n->i = i;
+  if (*il == NULL)
+    *il = n;
+  else {
+    for (; (*il)->next; (*il) = (*il)->next) ;
+    (*il)->next = n;
+  }
+}
+
 /* 
  * 3rd party tunneling per APRS spec 1.01 Ch. 17
  *
