@@ -61,11 +61,11 @@
 #include <time.h>
 #include <signal.h>
 
-#include <linux/ax25.h>
-#include <linux/rose.h>
-
-#include <netax25/axconfig.h>
+#include <netax25/kernel_ax25.h>
+#include <netax25/kernel_rose.h>
+#include <netax25/ax25.h>
 #include <netax25/axlib.h>
+#include <netax25/axconfig.h>
 
 #ifndef HAVE_LIBAX25_EXTENSIONS
 #include "libax25ext.h"
@@ -99,7 +99,7 @@ char *infofile = "/var/ax25/aprsmon.info";
 #endif
 
 #ifndef VERSION
-#define VERSION "$Revision: 1.15 $"
+#define VERSION "$Revision: 1.1.1.1 $"
 #endif
 
 char *title = "Live data from Linux";
